#include "board.h"
#include "util.h"
#include <string>
#include <ostream>
#include <stdexcept>
#include <stdlib.h>
#include <list>
#include <algorithm>
#include "robot.h"

using std::string;
using std::list;

//------------- board_type -----------------------------------

board_type board;

bool
board_type::read(int fd) {
    string size;
    if (!readline(fd, size)) {
        return false;
    }

    int w, h;
    int rc = sscanf(size.c_str(), "%d %d", &w, &h);
    if (rc != 2) {
        throw std::runtime_error("bad board size");
    }
    
    width = w + 2;
    height = h + 2;

    _tiles.resize(width * height, wall);
    _self_dist.resize(width * height);

    for (int i = 1; i <= h; i++) {
        string line;
        if (!readline(fd, line)) {
            return false;
        }

        for (int j = 1; j <= w; j++) {
            tile_type tile = (tile_type) line.at(j - 1);
            _tiles.at(i * width + j) = tile;
            if (tile == home_base) {
                home_bases.push_back(loc_type(j, i));
            }
        }
    }

    home_base_explored.resize(home_bases.size(), false);
//     fill(home_base_explored.begin(),
//          home_base_explored.begin() + home_base_explored.size() / 2 - 3, true);
    return true;
}

void
board_type::print(ostream &out) {
    out << width << ' ' << height << '\n';
    for (int i = 0; i < height; i++) {
        for (int j = 0; j < width; j++) {
            out << (char) _tiles.at(i * width + j);
        }
        out << '\n';
    }
}

static void printdist(int w, int h, 
	const vector<int>& sdist, const vector<bool>& visit)
{
    for(int y=0; y<h; y++)
    {
	for(int x=0; x<w; x++)
	{
	    int idx = y*w+x;
	    if(sdist[idx] > 10000000)
		printf("XX%c ",visit[idx]?'@':'.');
	    else
	        printf("%2.2d%c ", sdist[idx], visit[idx]?'@':'.');
	}
	printf("\n");
    }
}

void
board_type::get_neighbours(const loc_type& loc, tile_type dst[]) {
    get_neighbours(loc.y() * width + loc.x(), dst);
}

void
board_type::get_neighbours(int curidx, tile_type dst[]) {
    const int num_neigh = 4;
    int neigh_dx[] = {-1, +1,  0,  0};
    int neigh_dy[] = { 0,  0, -1, +1};

    for(int i=0; i<num_neigh; i++) {
        int neigh_idx = curidx + neigh_dy[i]*width+neigh_dx[i];
        dst[i] = _tiles[neigh_idx];
    }
}

void
board_type::get_neighbour_indexes(const loc_type &loc, int dst[]) {
    const int num_neigh = 4;
    int neigh_dx[] = {-1, +1,  0,  0};
    int neigh_dy[] = { 0,  0, -1, +1};

    int curidx = loc.y() * width + loc.x();
    for(int i=0; i<num_neigh; i++)
    {
        dst[i] = curidx + neigh_dy[i]*width+neigh_dx[i];
    }
}

int
board_type::get_scariness(const loc_type &loc, vector<bool> &robots) {
    // return the scariness of location
    int result = 0;

    // if there are no other robots, scariness does not apply
    if (live_robot_count() <= 1) {
        return 0;
    }
    
    tile_type neighbours[4];
    get_neighbours(loc, &neighbours[0]);

    for(int i=0; i<4; i++) {
        switch(neighbours[i]) {
        case space:
            break;
        case wall:
            break;
        case water:
            result += 2;
            break;
        case home_base:
            break;
        default:
            break;
        }

        // if robot on square, add 3
        int idx = loc.x() + loc.y() * width;
        if(robots[idx]) {
            result += 3;
        }
    }
    
    return result;
}

const int VISIBILITY = 10;

void
board_type::compute_from_self(const loc_type& selfloc)
{
    list<loc_type> locs[2];
    int current = 0;
    int next = 1;
    int curdist = 0;

    // start at selfloc
    locs[current].push_back(selfloc);
    // init array
    memset(&_self_dist.at(0), 0x0C, sizeof(int)*_self_dist.size());
    // temp array to see if location has been visited
    vector<bool> visited;
    visited.resize(width*height, false);
    visited.at(selfloc.y()*width+selfloc.x()) = true;

    // indicating robot present
    vector<bool> robot_present;
    robot_present.resize(width*height, false);
    for (robot_map::iterator it = robots.begin(); it != robots.end(); ++it) {
        robot_type *r = (*it).second;
        if (r == self || r->is_dead()) {
            continue;
        }
        
        int ridx = r->location.x() + r->location.y() * width;
        robot_present[ridx] = true;

        // neighbours too
        int indexes[4];
        get_neighbour_indexes(r->location, indexes);
        for(int i=0; i<4; i++) {
            robot_present[indexes[i]] = true;
        }
    }

    // neighbors of a tile
    const int num_neigh = 4;
    int neigh_dx[] = {-1, +1,  0,  0};
    int neigh_dy[] = { 0,  0, -1, +1};

    int selfidx = selfloc.y() * width + selfloc.x();
    _self_dist[selfidx] = curdist;

    while(locs[current].size() > 0)
    {
	locs[next].clear();
	// fill current with curdist
	for(list<loc_type>::const_iterator it=locs[current].begin();
		it!=locs[current].end(); it++)
	{
	    int curidx = (*it).y() * width + (*it).x();
            
	    // add neighbors
	    for(int i=0; i<num_neigh; i++)
	    {
		int neigh_idx = curidx + neigh_dy[i]*width+neigh_dx[i];
                loc_type neigh_loc ((*it).x()+neigh_dx[i],
                                    (*it).y()+neigh_dy[i]);
                int neigh_value = _self_dist[curidx] +
                    get_scariness(neigh_loc, robot_present) + 1;

                // ignore other tiles
		if(_tiles[neigh_idx] == space
                   || _tiles[neigh_idx] == home_base) {

                    // cutoff -- if the distance is > VISIBILITY, then
                    // it is so far away we don't have to include it
                    // again.
                    bool cutoff = false;
                    if(_self_dist[curidx] > VISIBILITY)
                        cutoff = true;

                    if(!visited[neigh_idx]
                       || ( _self_dist[neigh_idx] > neigh_value && !cutoff))
                        {
                            _self_dist[neigh_idx] = neigh_value;
                            visited[neigh_idx] = true;
                            locs[next].push_back(neigh_loc);
                        }
                }
            }
	}
	
	// swap lists
	current ^= 1;
	next ^= 1;
	curdist++;
    }
    //printdist(width, height, _self_dist, visited);
}

dir_type 
board_type::to_go(const loc_type& loc)
{
    int oldx = loc.x();
    int oldy = loc.y();
    int newx=oldx, newy=oldy;
    if(newx == self->location.x() && newy == self->location.y())
    {
	return north; 
    }
    // neighbors of a tile
    const int num_neigh = 4;
    int neigh_dx[] = {-1, +1,  0,  0};
    int neigh_dy[] = { 0,  0, -1, +1};
    
    /// dir to go to to get to previous spot.
    dir_type neigh_dir[] = {east, west, north, south};
    dir_type last_dir = north;

    int disttogo = _self_dist.at(newy*width+newx);
    if(disttogo > 10000000) return north;

    while(newx!=self->location.x() || newy!=self->location.y())
    {
        // no longer valid -- distance no longer 'accurate' since it
        // takes scariness into account.
        
//  	if(disttogo < 0) 
//  	{
//  	    assert(false);
//  	    return last_dir;
//  	}
	oldx = newx;
	oldy = newy;
//	disttogo--;

        // array of all possible closest locations
        int possibilities[num_neigh];
        int possibility_idx = 0;

	/// which direction is smallest
        int min = 10000000;
	for(int i=0; i<num_neigh; i++)
	{
            int idx = (newy+neigh_dy[i])*width + newx+neigh_dx[i];
	    if(_self_dist.at(idx) < min)
	    {
                min = _self_dist.at(idx);
	    }
	}

        // add all smallest to possibilities
	for(int i=0; i<num_neigh; i++)
	{
            int idx = (newy+neigh_dy[i])*width + newx+neigh_dx[i];
	    if(_self_dist.at(idx) == min)
	    {
                possibilities[possibility_idx++] = i;
	    }
	}

        // last_dir = random uit lijst. if possibility_idx == 0, we get
        // north which was the default for last_dir anyway.
        std::random_shuffle(&possibilities[0],
                            &possibilities[possibility_idx]);
        last_dir = neigh_dir[possibilities[0]];
        newy += neigh_dy[possibilities[0]];
        newx += neigh_dx[possibilities[0]];
    }
    return last_dir;
}

//------------- loc_type -----------------------------------

int
loc_type::manhattan_distance(const loc_type& other) const
{
    return abs(other.x() - _x) + abs(other.y() - _y);
}

int
loc_type::distance(const loc_type& other) const
{
    if (manhattan_distance(other) <= 1) {
        return manhattan_distance(other);
    } else if (*this == self->location) {
        return board.dist_from_self(other);
    } else if (other == self->location) {
        return board.dist_from_self(*this);
    } else {
        int d1 = abs(other.x() - _x) + abs(other.y() - _y);
        int d2 = board.dist_from_self(*this) + board.dist_from_self(other);
        return (3*d1 + d2) / 4;
    }
}
