// -*- C++ -*-

#ifndef BOARD_H
#define BOARD_H

#include <ostream>
#include <vector>
#include <stdexcept>
#include "util.h"

using std::ostream;
using std::vector;

enum dir_type {
    north = 'N',
    east  = 'E',
    south = 'S',
    west  = 'W'
};

enum tile_type {
    space = '.',
    wall  = '#',
    water = '~',
    home_base = '@'
};

static inline dir_type
left_dir(dir_type d) {
    switch (d) {
    case north:
        return west;
    case west:
        return south;
    case south:
        return east;
    case east:
        return north;
    }
}

static inline dir_type
right_dir(dir_type d) {
    switch (d) {
    case north:
        return east;
    case west:
        return north;
    case south:
        return west;
    case east:
        return south;
    }
}

class loc_type {
private:
    int _x;
    int _y;
public:
    loc_type() : _x(0), _y(0) {}
    loc_type(int ix, int iy) : _x(ix), _y(iy) {}
    loc_type(loc_type const &rhs) : _x(rhs._x), _y(rhs._y) {}

    loc_type const &operator=(loc_type const &rhs) {
        if (this != &rhs) {
            _x = rhs._x;
            _y = rhs._y;
        }
        return *this;
    }

    int x() const { return _x; }
    int y() const { return _y; }

    inline int index() const;
    inline loc_type move(dir_type d) const;
    inline tile_type operator*() const;
    /// get estimate of distance to other location
    int distance(const loc_type& other) const;
    int manhattan_distance(const loc_type& other) const;
};

inline bool operator==(loc_type const &lhs, loc_type const &rhs) {
    return lhs.x() == rhs.x() && lhs.y() == rhs.y();
}
inline bool operator!=(loc_type const &lhs, loc_type const &rhs) {
    return lhs.x() != rhs.x() || lhs.y() != rhs.y();
}

inline bool operator<(loc_type const &lhs, loc_type const &rhs) {
    return (lhs.x() < rhs.x() || (lhs.x() == rhs.x() && lhs.y() < rhs.y()));
}

inline ostream &operator<<(ostream &out, loc_type const &rhs) {
    return out << "(" << rhs.x() << ", " << rhs.y() << ")";
}

class board_type {
private:
    vector<tile_type> _tiles;
    vector<int> _self_dist;
public:
    static stat package_stat;
    
    int width;
    int height;
    vector<loc_type> home_bases;
    vector<bool> home_base_explored;
    
    board_type() {};
    ~board_type() {};

    bool read(int fd);

    void print(ostream &out);

    tile_type operator[](loc_type const &loc) {
        return _tiles.at(loc.y() * width + loc.x());
    }

    /// fill matrix for self_distance and path 
    void compute_from_self(const loc_type& selfloc);
    /// get accurate distance from self to location
    int dist_from_self(const loc_type& loc) const {
	return _self_dist.at(loc.y() * width + loc.x());
    }
    /// get direction for step towards target
    dir_type to_go(const loc_type& loc);

    // get distance w/scariness factored in
    int get_scariness(const loc_type& loc, vector<bool> &robots);

    // neighbours
    void get_neighbours(const loc_type& loc, tile_type dst[]);
    void get_neighbours(int loc, tile_type dst[]);
    void get_neighbour_indexes(const loc_type &loc, int dst[]);
};

extern board_type board;

inline loc_type
loc_type::move(dir_type d) const {
    switch (d) {
    case east:
        return loc_type(_x + 1, _y);
    case north:
        return loc_type(_x, _y + 1);
    case west:
        return loc_type(_x - 1, _y);
    case south:
        return loc_type(_x, _y - 1);
    default:
        throw std::runtime_error("bad direction");
    }
}

inline tile_type
loc_type::operator*() const {
    return board[*this];
}

inline int
loc_type::index() const {
    return _y * board.width + _x;
}

#endif // BOARD_H
