#include "evaluate.h"
#include "timer.h"
#include "util.h"
#include "board.h"
#include "robot.h"
#include "package.h"

#include <iostream>
#include <cstdlib>
#include <stdexcept>
#include <sstream>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <limits>

using namespace std;

struct client_config {
    client_config() : server_name(0), server_port(0), server_addr(0),
                      server_socket(-1), score(0), cpu_seconds_per_turn(1.0),
                      x(0), y(0) {}
    char const *server_name;
    char const *server_port;
    addrinfo *server_addr;
    int server_socket;
    int score;
    float cpu_seconds_per_turn;
    int x, y;
};

static bool
parse_result(client_config &config) {
    string result;
more_results:
    if (!readline(config.server_socket, result)) {
        cerr << "Connection closed.\n";
        return false;
    }

    istringstream input(result);
    int current_robot_id = 0;
    robot_type *current = 0;
    char ch;
    while (input.get(ch)) {
        if (isspace(ch))
            continue;

        if (ch == '#') {
            input >> current_robot_id;
            current = get_robot(current_robot_id);
        } else {
            input.unget();
            string cmd;
            input >> cmd;
            if (cmd == "Robot") {
                int rid;
                input >> rid;
                goto more_results;
            } else if (cmd == "***") {
                goto more_results;
            }
                
            assert(current);
            current->not_dead();
            switch (cmd[0]) {
            case 'N': case 'E': case 'S': case 'W':
                assert(current);
//                 cerr << current_robot_id << " moved " << cmd << '\n';;
                current->move((dir_type) cmd[0]);
                break;
            case 'P': {
                int pid;
                input >> pid;
//                 cerr << current_robot_id << " picked " << pid << '\n';
                get_package(pid, current); // automagically updates carrier
                break;
            }
            case 'D': {
                int pid;
                input >> pid;
//                 cerr << current_robot_id << " dropped " << pid << '\n';
                // FIXME
                package_type *p = get_package(pid, current);
                p->set_location(current->location);
                if (current == self && p->destination && p->location() &&
                    *p->destination == *p->location())
                {
                    config.score += p->weight();
                    p->set_delivered();
                }
                break;
            }
            case 'X': {
                int x, y;
                input >> x >> cmd >> y;
//                 cerr << current_robot_id << " is at " << x << ", " << y << '\n';
                current->location = loc_type(x, y);
                break;
            }
            default:
                throw std::runtime_error("bad command");
            }
        }
    }

    return true;
}

static void
print_status(ostream &out) {
    print_robots(out);
    print_packages(out);
}

static bool
parse_packages(client_config const &config, package_vector &available) {
    string packages;
    if (!readline(config.server_socket, packages)) {
        cerr << "Connection closed.\n";
        return false;
    }

    available.clear();
    
    istringstream input(packages);
    int pid, dest_x, dest_y, weight;
    while (input >> pid >> dest_x >> dest_y >> weight) {
        package_type *package = get_package(
            pid, loc_type(dest_x, dest_y), weight, self->location);
        assert(package);
        available.insert(package);
    }

    return true;
}

int
main(int argc, char **argv) {
    int rc;
    
    timer main_timer;
    main_timer.start();

    srand(time(0));
    
    // Packages have a weight of at least one... This sets the
    // starting average to 1.0 and helps keep the calculated average
    // down a little.
    package_type::weight_stat.add_double(1.0);
    
    if (argc < 3) {
        cerr << "Usage: " << argv[0] << " server_name server_port\n";
        return EXIT_FAILURE;
    }

    client_config config;
    config.server_name = argv[1];
    config.server_port = argv[2];

    cerr << "Server: " << config.server_name
         << ", port: " << config.server_port << "\n";

    for (int i = 3; i < argc; ++i) {
        if (strcmp(argv[i], "-s") == 0) {
            config.cpu_seconds_per_turn = atof(argv[i + 1]);
            ++i;
        } if (strcmp(argv[i], "-l") == 0) {
            config.x = atoi(argv[i + 1]);
            config.y = atoi(argv[i + 2]);
            i += 2;
        }
    }
    
    // Look up server.
    addrinfo hint;
    hint.ai_flags = 0;
    hint.ai_family = PF_INET;
    hint.ai_socktype = SOCK_STREAM;
    hint.ai_protocol = 0;
    rc = getaddrinfo(
        config.server_name, config.server_port, &hint, &config.server_addr);
    if (rc != 0) {
        cerr << "getaddrinfo: " << gai_strerror(rc) << "\n";
        return EXIT_FAILURE;
    }

    config.server_socket = socket(config.server_addr->ai_family,
                                  config.server_addr->ai_socktype,
                                  config.server_addr->ai_protocol);
    if (config.server_socket < 0) {
        cerr << "socket: " << strerror(errno) << '\n';
        return EXIT_FAILURE;
    }

    rc = connect(config.server_socket,
                 config.server_addr->ai_addr,
                 config.server_addr->ai_addrlen);
    if (rc < 0) {
        cerr << "connect: " << strerror(errno) << '\n';
        return EXIT_FAILURE;
    }

    size_t moves = 1;
    try {
        if (config.x == 0) {
            sendline(config.server_socket, "Player");
        } else {
            ostringstream cmd;
            cmd << "Login " << config.x << ' ' << config.y;
            sendline(config.server_socket, cmd.str());
        }

        board.read(config.server_socket);
        
        if (!read_robot(config.server_socket)) {
            cerr << "Cannot read robot.\n";
            return EXIT_FAILURE;
        }
        
        parse_result(config);

        while (1) {
            timer command_timer;
            command_timer.start();

            package_vector available;
            if (!parse_packages(config, available))
                break;

//             print_status(cout);

//             cout << "#" << self->id << ", capacity " << self->capacity() << endl;
            cout << "Tuning: " << tuning << endl;
            
            string command = evaluate(available);
            sendline(config.server_socket, command);
            ++moves;

            for (robot_map::iterator it = robots.begin();
                 it != robots.end();
                 ++it)
            {
                robot_type *robot = (*it).second;
                robot->before_results();
            }
            
            if (!parse_result(config))
                break;

            command_timer.stop();
            main_timer.stop();
            float used_cpu_time_this_turn = command_timer.cpu_time();
            float used_cpu_time_per_turn = (main_timer.cpu_time() / moves);
            tuning = ((0.95 * config.cpu_seconds_per_turn) /
                      (((9.0 * used_cpu_time_per_turn
                         + used_cpu_time_this_turn)
                        / 10.0)
                       + 0.01));
            if (tuning > 1000.0) {
                tuning = 1000.0;
            } else if (tuning <= 0.0001) {
                tuning = 0.0001;
            }
            cout << "#" << moves
                 << " Command: " << command_timer.cpu_time()
                 << " seconds, " << used_cpu_time_per_turn
                 << " seconds avg/turn." << endl;

            //sleep(1);
        }
    } catch (std::runtime_error &ex) {
        cerr << "Caught: " << ex.what() << '\n';
        return EXIT_FAILURE;
    } catch (...) {
        cerr << "Caught unknown exception\n";
        return EXIT_FAILURE;
    }

    main_timer.stop();
    cout << "CPU time: " << main_timer.cpu_time()
         << " for " << moves << " moves ("
         << main_timer.cpu_time() / moves << " seconds/move).\n";
    cout << config.score << "\n";
    
    shutdown(config.server_socket, SHUT_RDWR);
    freeaddrinfo(config.server_addr);
    return EXIT_SUCCESS;
}
