#include "evaluate.h"
#include <algorithm>
#include "robot.h"
#include "board.h"
#include "timer.h"
#include <vector>
#include <sstream>
#include <iostream>
#include <limits>
#include <set>

using namespace std;

float tuning = 1.0;

static size_t
get_eval_limit(size_t last_eval_limit, float factor) {
    float new_eval_limit = last_eval_limit * factor;
    size_t eval_limit = (size_t) new_eval_limit;
    if (new_eval_limit > numeric_limits<size_t>::max()) {
        eval_limit = numeric_limits<size_t>::max();
    } else if (new_eval_limit < 1) {
        eval_limit = 1;
    }
    return eval_limit;
}

static float
distance_value_mod(int distance, int money) {
    return min(0.5, 1.0 - (float(distance) / float(money + 5)));
}

dir_type
action::direction() const {
    return board.to_go(location);
}

string
action::to_string() const {
    ostringstream tmp;
    tmp << bid << ' ';
    if (action_type == move) {
        tmp << "Move " << (char) board.to_go(location);
    } else {
        tmp << (action_type == pick ? "Pick" : "Drop");
        for (package_vector::const_iterator it = packages.begin();
             it != packages.end();
             ++it)
        {
            tmp << ' ' << (*it)->id;
        }
    }
    return tmp.str();
}

/// static sort operator
static bool sortpred(const action& a, const action& b)
{
    return (a.value > b.value);
}

void
fill_retrieve(vector<action>& possible, package_vector const &available);

int
determine_bid(int money, float percentage) {
    return min(
        self->money(),
        max(2,
            int(money * percentage * (0.5 + rand()/(RAND_MAX+1.0f)))));
}

static string
modify_for_survival(action &a) {
    vector<bool> robot_present;
    robot_present.resize(board.width*board.height, false);
    for (robot_map::iterator it = robots.begin(); it != robots.end(); ++it) {
        robot_type *r = (*it).second;
        if (r == self || r->is_dead()) {
            continue;
        }
        
        int ridx = r->location.x() + r->location.y() * board.width;
        robot_present[ridx] = true;
    }
        
    int indices[4];
    tile_type types[4];
    dir_type neigh_dir[] = {west, east, south, north};
    int opp_dir[] = {1, 0, 3, 2};
    
    board.get_neighbours(self->location, types);
    board.get_neighbour_indexes(self->location, indices);
    for (int i = 0; i < 4; ++i) {
        if (types[i] == water || types[i] == wall)
            continue;
        
        if (robot_present[indices[i]]) {
            tile_type enemy_types[4];
            board.get_neighbours(indices[i], enemy_types);
            if (enemy_types[i] == water) {
                action push(action::move, "push enemy");
                push.value = numeric_limits<float>::max();
                push.location = self->location.move(neigh_dir[i]);
                push.bid = determine_bid(self->money(), 0.1);
                return push.to_string();
            } else if (types[opp_dir[i]] == water) {
                action run(action::move, "back to water");
                run.value = numeric_limits<float>::max() / 2.0;
                run.location = a.location;
                run.bid = determine_bid(self->money(), 0.2);
                return run.to_string();
            }
        }
    }

    if (a.action_type == action::move) {
        dir_type dir = a.direction();
        loc_type forward = self->location.move(dir);
        loc_type forward_left = forward.move(left_dir(dir));
        loc_type forward_right = forward.move(right_dir(dir));
        
        if ((*forward_left == water && robot_present[forward_right.index()])
            || (*forward_right == water && robot_present[forward_left.index()]))
        {
            a.bid = -determine_bid(self->money(), 0.2);
        }
    }
    
    return a.to_string();
}

string
evaluate(package_vector &available) {
    vector<action> possible;

    // compute all paths from self
    board.compute_from_self(self->location);

    // fill
    fill_deliver(possible);
    fill_explore(possible);
    fill_retrieve(possible, available);

    // sort
    timer sort_timer;
    sort_timer.start();
    stable_sort(possible.begin(), possible.end(), sortpred);
    sort_timer.stop();
//    cerr << "evaluate: sort CPU: " << sort_timer.cpu_time() << endl;
    
//     for (vector<action>::iterator it = possible.begin(); it != possible.end();
//          ++it)
//     {
//         cerr << "action " << (*it).reason
//              << ", " << (*it).value
//              << " " << (*it).to_string() << '\n';
//     }
    
    // get best action
    if(possible.size() > 0) {
        action &a = possible.at(0);
//         cerr << "action reason: " << a.reason << ", score: " << a.value << endl;
        self->used_money(abs(a.bid));
        return modify_for_survival(a);
    }
    // never here
    assert(false);
    return "1 Move N";
}

typedef pair<float, set<package_type *> > package_set_value;
typedef map<package_type const *, package_set_value> package_value_map;

static float max_package_value;
static float const log_2 = log(2.0f);

void
update_package_map(package_value_map &possible, float current_value,
                   set<package_type *> &processed,
                   package_type *current)
{
//     cout << "pkg: " << *current << ", value: " << current_value
//          << ", #deps: " << processed.size() << endl;
    if (possible[current].first < current_value)
        possible[current] = package_set_value(current_value, processed);
}

void
generate_iterative_command(vector<action>& possible,
                           package_value_map &package_values,
                           vector<package_type *> &packages,
                           package_type *current,
                           size_t &eval_count, size_t eval_limit)
{
    set<package_type *> carry_set;
    int available_capacity = self->max_capacity();
    for (vector<package_type *>::iterator packet_it = packages.begin();
         packet_it != packages.end();
         ++packet_it)
    {
        ++eval_count;
        if (eval_count > eval_limit)
            goto generate_commands;
        
        package_type *package = *packet_it;
        if (package->weight() > available_capacity)
            continue;
        
        set<package_type *> &dependents = package_values[package].second;

        // Add any extra dependent packages.
        int current_capacity = available_capacity;
        set<package_type *> current_set;
        for (set<package_type *>::iterator it = dependents.begin();
             it != dependents.end()
                 && current_capacity >= 0;
             ++it)
        {
            ++eval_count;
            if (eval_count > eval_limit)
                goto generate_commands;
        
            package_type *p = *it;
            if (carry_set.find(p) == carry_set.end()) {
                current_capacity -= p->weight();
                current_set.insert(p);
            }
        }

        if (current_capacity >= 0) {
            carry_set.insert(current_set.begin(), current_set.end());
            available_capacity = current_capacity;
        }
    }

generate_commands:
    set<package_type *> to_drop_set;
    for (package_vector::iterator it = self->packages.begin();
         it != self->packages.end();
         ++it)
    {
        package_type *p = *it;
        if (carry_set.find(p) == carry_set.end()) {
            to_drop_set.insert(p);
        }
    }

    if (to_drop_set.size() > 0) {
        /* gen drop cmd */
        action drop(action::drop, "drop packages to free up capacity");
        drop.value = package_values[current].first;
        drop.packages.insert(to_drop_set.begin(),
                             to_drop_set.end());
        possible.push_back(drop);
        return;
    }
    
    set<package_type *> to_pick_set;
    for (set<package_type *>::iterator it = carry_set.begin();
         it != carry_set.end();
         ++it)
    {
        package_type *p = *it;
        if (self->packages.find(p) == self->packages.end()) {
            to_pick_set.insert(p);
        }
    }
    
    if (to_pick_set.size() > 0) {
        /* gen pick cmd */
        action pick(action::pick, "pick up packets for delivery");
        pick.value = package_values[current].first;
        pick.packages.insert(to_pick_set.begin(),
                             to_pick_set.end());
        possible.push_back(pick);
        return;
    }
    
    /* gen move cmd to current */
    action move(action::move, "iterative smart genius delivery");
    move.value = package_values[current].first;
    move.location = *current->destination;
    possible.push_back(move);
}

void
find_iterative(package_value_map &possible,
               vector<package_type *> const &available,
               size_t &eval_count, size_t eval_limit, size_t max_depth,
               loc_type const &loc,
               float current_value, set<package_type *> &processed,
               package_type *current, int current_capacity, int current_money)
{
    size_t depth_mod = 0;
    
//     if (current != 0) {
//         float best_value = possible[current].first;
    
//         if (max_package_value == 0) {
//             depth_mod = 0;
//         } else {
//             depth_mod = size_t(
//                 max(0.0f,
//                     min(float(available.size()),
//                         -log(best_value / max_package_value) / log(1.5f))));
//         }
// //         if (depth_mod != 0) {
// //             cerr << "depth_mod: " << depth_mod
// //                  << ", best_value: " << best_value
// //                  << ", max_package_value: " << max_package_value
// //                  << endl;
// //         }
//     }
    
    if (eval_count > eval_limit || max_depth <= depth_mod) {
        update_package_map(possible, current_value, processed, current);
        return;
    }
    
    --max_depth;
    for (vector<package_type *>::const_iterator it = available.begin();
         it != available.end(); ++it)
    {
        ++eval_count;
        if (eval_count > eval_limit)
            return;
        
        package_type *p = *it;
        if (processed.find(p) == processed.end()
            && p->destination
            && current_capacity + p->weight() <= self->max_capacity()
            && (current == 0
                || p->weight() < current->weight()
                || (p->weight() == current->weight()
                    && current->id < p->id)))
        {
            int penalty = p->carrier() == self ? 0 : 2;
            int distance = loc.distance(*p->destination) + penalty;
            float value = p->get_value(loc, penalty);
            value *= distance_value_mod(distance, current_money);
            processed.insert(p);
            find_iterative(
                possible, available,
                eval_count, eval_limit, max_depth,
                *p->destination,
                current_value + value,
                processed, current ? current : p,
                current_capacity + p->weight(),
                current_money - distance);
            processed.erase(p);
        }
    }
}

package_value_map *package_values;

bool
package_value_pred(package_type const *lhs, package_type const *rhs) {
    assert(lhs);
    assert(rhs);
    return (*package_values)[lhs].first > (*package_values)[rhs].first;
}

void
fill_retrieve(vector<action>& possible, package_vector const &available) {
    static size_t last_eval_limit = 1000;
    size_t eval_limit = get_eval_limit(last_eval_limit, tuning);
//     cerr << "fill_retrieve: eval_limit = " << eval_limit << ", " << last_eval_limit << endl;
    if (eval_limit < last_eval_limit)
        last_eval_limit = eval_limit;
    size_t eval_count = 0;

    vector<package_type *> packages;
    packages.insert(packages.end(),
                    self->packages.begin(), self->packages.end());
    packages.insert(packages.end(),
                    available.begin(), available.end());

    size_t eval_count2 = 0;
    
    max_package_value = 0.0;
    package_values = new package_value_map;
    for (size_t max_depth = 1;
         max_depth <= packages.size();
         ++max_depth)
    {
        ++eval_count;
        if (eval_count > eval_limit)
            break;
        
        set<package_type *> processed;
        find_iterative(
            *package_values, packages, eval_count, eval_limit, max_depth,
            self->location, 0.0, processed, 0, 0,
            self->money());

//         cerr << "fill_retrieve: #pkgs = " << packages.size()
//              << ", max_depth = " << max_depth
//              << ", eval_count = " << eval_count
//              << endl;

//         timer sort_timer;
//         sort_timer.start();
        stable_sort(packages.begin(), packages.end(), package_value_pred);
//         sort_timer.stop();
//         cerr << "find_retrieve: sort CPU: " << sort_timer.cpu_time() << endl;

//         for (vector<package_type *>::iterator it = packages.begin();
//              it != packages.end();
//              ++it)
//         {
//             package_type *p = *it;
//             cout << "max " << max_depth
//                  << " package " << *p << " value " << (*package_values)[p].first << endl;
//         }

        // Generate commands.
        if (packages.size() > 0) {
            max_package_value = (*package_values)[packages.at(0)].first;
            generate_iterative_command(
                possible,
                *package_values,
                packages,
                packages.at(0),
                eval_count2, eval_limit);
        }
    }

    delete package_values;
    
    if (eval_count > last_eval_limit) {
        last_eval_limit = eval_count;
    } else if (eval_count2 > last_eval_limit) {
        last_eval_limit = eval_count2;
    } else {
//         cout << "find_retrieve: limit not reached.\n";
    }
}


// O(#home_bases * #packages)
void
fill_explore(vector<action>& possible) {
    static size_t last_eval_limit = 1000;
    size_t eval_limit = get_eval_limit(last_eval_limit, tuning);
    if (eval_limit < last_eval_limit)
        last_eval_limit = eval_limit;
    size_t eval_count = 0;
//     cerr << "fill_explore: eval_limit = " << eval_limit << endl;
    vector<bool>::iterator it_expl=board.home_base_explored.begin();
    for(vector<loc_type>::const_iterator it=board.home_bases.begin();
        it!=board.home_bases.end(); ++it, ++it_expl)
    {
        const loc_type& loc = *it;
        //int distance = self->location.distance(loc);
        if (self->location == loc) {
            (*it_expl) = true;
            continue;
        }
        
        int distance = board.dist_from_self(loc);
        action a(action::move, "explore home base");
        a.value = 1.0 / float(distance + 2);
        a.value *= distance_value_mod(distance + 2, self->money());

        if(*it_expl) 
        {
            // value is sum of package values at homebase
            a.value = 0.0;
            int current_capacity = self->capacity();
            for(packages_by_loc_type::const_iterator p_it=
                    packages_by_loc.lower_bound(loc); 
                p_it != packages_by_loc.upper_bound(loc); ++p_it)
            {
                package_type& pack = *(*p_it).second;
                if(pack.weight() <= current_capacity)
                {
                    current_capacity -= pack.weight();
                    int todeliver = (board.width+board.height+2)/3;
                    if(pack.destination)
                        todeliver = loc.distance(*pack.destination);
                    a.value += (float(pack.weight()) / float(
                                    distance + todeliver + 2)
                                * distance_value_mod(distance + todeliver + 2,
                                                     self->money()));
                }
            }
            if(a.value == 0.0) continue; // seen homebase, but not interesting
            a.reason = "return to home base";
        }

        ++eval_count;
        if (eval_count > eval_limit) {
            break;
        }

        a.location = loc;
        possible.push_back(a);
//      cout << "Homebase moving value "<<a.value<<endl;
//      cout << loc << endl;
    }

    if (eval_count > last_eval_limit)
        last_eval_limit = eval_count;

    // add a random action (just in case)
    float rnd = rand()/(RAND_MAX+1.0f);
    dir_type dir = north;
    if(rnd <= 0.25)
        dir = east;
    else if(rnd <= 0.5)
        dir = west;
    else if(rnd <= 0.75)
        dir = south;
    action duh(action::move, "random");
    duh.value = 1e-5f;
    duh.location = self->location.move(dir);

    // check killingness
    loc_type loc = self->location;
    loc.move(dir);
    if(*duh.location == water) 
        duh.action_type = action::drop;

    possible.push_back(duh);
//     cout << "Random moving value "<<duh.value<<duh.todo<<endl;
}


// O(#packages)
void
fill_deliver(vector<action>& possible)
{
    // dropping packages most important
    action drop(action::drop, "deliver package");
    for(package_vector::const_iterator it=self->packages.begin();
        it!=self->packages.end(); ++it)
    {
        package_type& pack = *(*it);
        if(!pack.destination || *pack.destination == self->location)
        {
            drop.value = 10.0e9;
            drop.packages.insert(&pack);
        }
    }
    possible.push_back(drop);
}
