#include "package.h"
#include <algorithm>
#include <ostream>
#include <stdexcept>
#include <utility>
#include <iostream>

using namespace std;

package_map packages;
packages_by_loc_type packages_by_loc;

stat package_type::weight_stat;

package_type::package_type(int iid, robot_type *carrier)
    : id(iid), destination(0), _weight(0), _location(0), _carrier(0),
      delivered(false)
{
    set_carrier(carrier);
}

package_type::package_type(int iid, loc_type const &idest, int w, loc_type const &iloc)
    : id(iid), destination(new loc_type(idest)), _weight(new int(w)),
      _location(0), _carrier(0), delivered(false) {
    weight_stat.add_double(w);
    set_location(iloc);
}

void
package_type::print(ostream &out) {
    out << *this << '\n';
}

void
package_type::set_location(loc_type const &loc) {
    
    if (_carrier) {
        _carrier->remove(this);
        _carrier = 0;
    }
        
    if (_location) {
        packages_by_loc.erase(
            find(packages_by_loc.lower_bound(*_location),
                 packages_by_loc.upper_bound(*_location),
                 make_pair<const loc_type, package_type *>(*_location, this)));
        delete _location;
    }
    
    _location = new loc_type(loc);
    packages_by_loc.insert(make_pair(loc, this));

    assert((_carrier || _location)
           && !(_carrier && _location));
}
    
void
package_type::set_carrier(robot_type *robot) {
    assert(robot);

    if (_location) {
        packages_by_loc.erase(
            find(packages_by_loc.lower_bound(*_location),
                 packages_by_loc.upper_bound(*_location),
                 make_pair<const loc_type, package_type *>(*_location, this)));
        delete _location;
        _location = 0;
    }
        
    if (_carrier == robot) {
        return;
    }
        
    if (_carrier) {
        _carrier->remove(this);
        _carrier = 0;
    }

    _carrier = robot;
    _carrier->add(this);

    assert((_carrier || _location)
           && !(_carrier && _location));
}
    
ostream &
operator<<(ostream &out, package_type const &rhs) {
    out << rhs.id;
    if (rhs.destination) {
        out << ' ' << *rhs.destination;
    } else {
        out << " (unknown)";
    }
    if (rhs.real_weight()) {
        out << ' ' << *rhs.real_weight();
    } else {
        out << " (unknown)";
    }
    if (rhs.location()) {
        out << ' ' << *rhs.location();
    } else {
        out << " carrier: " << rhs.carrier();
    }
    return out;
}

package_type *
get_package(int id) {
    return (*packages.insert(make_pair(id, (package_type *) 0)).first).second;
}

package_type *
get_package(int id, robot_type *carrier) {
    package_type *result = get_package(id);
    if (result == 0) {
        result = packages[id] = new package_type(id, carrier);
    } else {
        result->set_carrier(carrier);
    }
    return result;
}

package_type *
get_package(int id, loc_type destination, int weight, loc_type location) {
    package_type *result = get_package(id);
    if (result == 0) {
        result = packages[id] = new package_type(
            id, destination, weight, location);
    } else {
        result->set_destination(destination);
        result->set_weight(weight);
        result->set_location(location);
    }
    return result;
}

void
print_packages(ostream &out) {
    out << "Packages:\n";
    for (package_map::iterator it = packages.begin(); it != packages.end(); ++it) {
        (*it).second->print(out);
    }
}

float
package_type::get_value() const
{
//     cout << "pack " << id << " loc / del"<<endl;
    if(!destination) return 0.0;
    if(delivered) return 0.0;
    if(_carrier && _carrier != self) return 0.0;
    int dist = board.dist_from_self(*destination);
//     cout << "pack " << id << " dist "<< dist<< endl;
    if(dist > 10000000) return 0.0;
    return float(weight()) / (float(dist + _carrier ? 1 : 3));
}

float
package_type::get_value(loc_type const &from, int penalty) const
{
//     cout << "pack " << id << " loc / del"<<endl;
    if(!destination) return 0.0;
    if(delivered) return 0.0;
    int dist = from.distance(*destination);
//     cout << "pack " << id << " dist "<< dist<< endl;
    if(dist > 10000000) return 0.0;
    float result = float(weight()) / (float(dist + penalty + 1));
    return result;
}
